@@ -28,7 +28,7 @@ import org.tukaani.xz.check.Check;
  * Unless you know what you are doing, don't use this class to decompress
  * standalone .xz files. For that purpose, use <code>XZInputStream</code>.
  *
- * <h4>When uncompressed size is known beforehand</h4>
+ * <h2>When uncompressed size is known beforehand</h2>
  * <p>
  * If you are decompressing complete XZ streams and your application knows
  * exactly how much uncompressed data there should be, it is good to try
@@ -41,6 +41,7 @@ import org.tukaani.xz.check.Check;
  */
 public class SingleXZInputStream extends InputStream {
     private InputStream in;
+    private final ArrayCache arrayCache;
     private final int memoryLimit;
     private final StreamFlags streamHeaderFlags;
     private final Check check;
@@ -93,12 +94,47 @@ public class SingleXZInputStream extends InputStream {
         this(in, -1);
     }
 
+    /**
+     * Creates a new XZ decompressor that decompresses exactly one
+     * XZ Stream from <code>in</code> without a memory usage limit.
+     * <p>
+     * This is identical to <code>SingleXZInputStream(InputStream)</code>
+     * except that this also takes the <code>arrayCache</code> argument.
+     *
+     * @param       in          input stream from which XZ-compressed
+     *                          data is read
+     *
+     * @param       arrayCache  cache to be used for allocating large arrays
+     *
+     * @throws      XZFormatException
+     *                          input is not in the XZ format
+     *
+     * @throws      CorruptedInputException
+     *                          XZ header CRC32 doesn't match
+     *
+     * @throws      UnsupportedOptionsException
+     *                          XZ header is valid but specifies options
+     *                          not supported by this implementation
+     *
+     * @throws      EOFException
+     *                          less than 12 bytes of input was available
+     *                          from <code>in</code>
+     *
+     * @throws      IOException may be thrown by <code>in</code>
+     *
+     * @since 1.7
+     */
+    public SingleXZInputStream(InputStream in, ArrayCache arrayCache)
+            throws IOException {
+        this(in, -1, arrayCache);
+    }
+
     /**
      * Creates a new XZ decompressor that decompresses exactly one
      * XZ Stream from <code>in</code> with an optional memory usage limit.
      * <p>
      * This is identical to <code>SingleXZInputStream(InputStream)</code>
-     * except that this takes also the <code>memoryLimit</code> argument.
+     * except that this also takes the <code>memoryLimit</code> argument.
      *
      * @param       in          input stream from which XZ-compressed
      *                          data is read
@@ -125,7 +161,47 @@ public class SingleXZInputStream extends InputStream {
      */
     public SingleXZInputStream(InputStream in, int memoryLimit)
             throws IOException {
-        this(in, memoryLimit, true, readStreamHeader(in));
+        this(in, memoryLimit, true);
+    }
+
+    /**
+     * Creates a new XZ decompressor that decompresses exactly one
+     * XZ Stream from <code>in</code> with an optional memory usage limit.
+     * <p>
+     * This is identical to <code>SingleXZInputStream(InputStream)</code>
+     * except that this also takes the <code>memoryLimit</code> and
+     * <code>arrayCache</code> arguments.
+     *
+     * @param       in          input stream from which XZ-compressed
+     *                          data is read
+     *
+     * @param       memoryLimit memory usage limit in kibibytes (KiB)
+     *                          or <code>-1</code> to impose no
+     *                          memory usage limit
+     *
+     * @param       arrayCache  cache to be used for allocating large arrays
+     *
+     * @throws      XZFormatException
+     *                          input is not in the XZ format
+     *
+     * @throws      CorruptedInputException
+     *                          XZ header CRC32 doesn't match
+     *
+     * @throws      UnsupportedOptionsException
+     *                          XZ header is valid but specifies options
+     *                          not supported by this implementation
+     *
+     * @throws      EOFException
+     *                          less than 12 bytes of input was available
+     *                          from <code>in</code>
+     *
+     * @throws      IOException may be thrown by <code>in</code>
+     *
+     * @since 1.7
+     */
+    public SingleXZInputStream(InputStream in, int memoryLimit,
+                               ArrayCache arrayCache) throws IOException {
+        this(in, memoryLimit, true, arrayCache);
     }
 
     /**
@@ -134,7 +210,7 @@ public class SingleXZInputStream extends InputStream {
      * and ability to disable verification of integrity checks.
      * <p>
      * This is identical to <code>SingleXZInputStream(InputStream,int)</code>
-     * except that this takes also the <code>verifyCheck</code> argument.
+     * except that this also takes the <code>verifyCheck</code> argument.
      * <p>
      * Note that integrity check verification should almost never be disabled.
      * Possible reasons to disable integrity check verification:
@@ -182,11 +258,59 @@ public class SingleXZInputStream extends InputStream {
      */
     public SingleXZInputStream(InputStream in, int memoryLimit,
                                boolean verifyCheck) throws IOException {
-        this(in, memoryLimit, verifyCheck, readStreamHeader(in));
+        this(in, memoryLimit, verifyCheck, ArrayCache.getDefaultCache());
+    }
+
+    /**
+     * Creates a new XZ decompressor that decompresses exactly one
+     * XZ Stream from <code>in</code> with an optional memory usage limit
+     * and ability to disable verification of integrity checks.
+     * <p>
+     * This is identical to
+     * <code>SingleXZInputStream(InputStream,int,boolean)</code>
+     * except that this also takes the <code>arrayCache</code> argument.
+     *
+     * @param       in          input stream from which XZ-compressed
+     *                          data is read
+     *
+     * @param       memoryLimit memory usage limit in kibibytes (KiB)
+     *                          or <code>-1</code> to impose no
+     *                          memory usage limit
+     *
+     * @param       verifyCheck if <code>true</code>, the integrity checks
+     *                          will be verified; this should almost never
+     *                          be set to <code>false</code>
+     *
+     * @param       arrayCache  cache to be used for allocating large arrays
+     *
+     * @throws      XZFormatException
+     *                          input is not in the XZ format
+     *
+     * @throws      CorruptedInputException
+     *                          XZ header CRC32 doesn't match
+     *
+     * @throws      UnsupportedOptionsException
+     *                          XZ header is valid but specifies options
+     *                          not supported by this implementation
+     *
+     * @throws      EOFException
+     *                          less than 12 bytes of input was available
+     *                          from <code>in</code>
+     *
+     * @throws      IOException may be thrown by <code>in</code>
+     *
+     * @since 1.7
+     */
+    public SingleXZInputStream(InputStream in, int memoryLimit,
+                               boolean verifyCheck, ArrayCache arrayCache)
+            throws IOException {
+        this(in, memoryLimit, verifyCheck, readStreamHeader(in), arrayCache);
     }
 
     SingleXZInputStream(InputStream in, int memoryLimit, boolean verifyCheck,
-                        byte[] streamHeader) throws IOException {
+                        byte[] streamHeader, ArrayCache arrayCache)
+            throws IOException {
+        this.arrayCache = arrayCache;
         this.in = in;
         this.memoryLimit = memoryLimit;
         this.verifyCheck = verifyCheck;
@@ -294,7 +418,8 @@ public class SingleXZInputStream extends InputStream {
                 if (blockDecoder == null) {
                     try {
                         blockDecoder = new BlockInputStream(
-                                in, check, verifyCheck, memoryLimit, -1, -1);
+                                in, check, verifyCheck, memoryLimit, -1, -1,
+                                arrayCache);
                     } catch (IndexIndicatorException e) {
                         indexHash.validate(in);
                         validateStreamFooter();
@@ -360,13 +485,48 @@ public class SingleXZInputStream extends InputStream {
     /**
      * Closes the stream and calls <code>in.close()</code>.
      * If the stream was already closed, this does nothing.
+     * <p>
+     * This is equivalent to <code>close(true)</code>.
      *
      * @throws  IOException if thrown by <code>in.close()</code>
      */
     public void close() throws IOException {
+        close(true);
+    }
+
+    /**
+     * Closes the stream and optionally calls <code>in.close()</code>.
+     * If the stream was already closed, this does nothing.
+     * If <code>close(false)</code> has been called, a further
+     * call of <code>close(true)</code> does nothing (it doesn't call
+     * <code>in.close()</code>).
+     * <p>
+     * If you don't want to close the underlying <code>InputStream</code>,
+     * there is usually no need to worry about closing this stream either;
+     * it's fine to do nothing and let the garbage collector handle it.
+     * However, if you are using {@link ArrayCache}, <code>close(false)</code>
+     * can be useful to put the allocated arrays back to the cache without
+     * closing the underlying <code>InputStream</code>.
+     * <p>
+     * Note that if you successfully reach the end of the stream
+     * (<code>read</code> returns <code>-1</code>), the arrays are
+     * automatically put back to the cache by that <code>read</code> call. In
+     * this situation <code>close(false)</code> is redundant (but harmless).
+     *
+     * @throws  IOException if thrown by <code>in.close()</code>
+     *
+     * @since 1.7
+     */
+    public void close(boolean closeInput) throws IOException {
         if (in != null) {
+            if (blockDecoder != null) {
+                blockDecoder.close();
+                blockDecoder = null;
+            }
+
             try {
-                in.close();
+                if (closeInput)
+                    in.close();
             } finally {
                 in = null;
             }
